/*
   Setup, Variable declarations and main loop of the program
*/

// SPI-Library
#include <SPI.h>

// i2c-Library
#include <Wire.h>

// Display-Libraries
#include <Adafruit_GFX.h>
#include <gfxfont.h>
#include <Adafruit_HX8357.h>

// Touch-Library
#include <Adafruit_STMPE610.h>

// ADC-Library
#include <Adafruit_ADS1015.h>


// Setup-Data for TFT-Display
#define TFT_DC 0  // Connect TFT_DC to D3
#define TFT_CS 2  // Connect TFT_CS to D4
#define TFT_RST -1 //
Adafruit_HX8357 tft = Adafruit_HX8357(TFT_CS, TFT_DC); // Create Display-Object
Adafruit_STMPE610 touch = Adafruit_STMPE610(); // Create TouchController-Object

// Setup-Data for the ADC on the analog board
Adafruit_ADS1015 ads;     /* Use thi for the 12-bit version */

const String SysVersion = "Version 0.3";

// Assign HEX-Values to color-names
#define BLACK 0x0000
#define BLUE 0x001F
#define RED 0xF800
#define GREEN 0x07E0
#define CYAN 0x07FF
#define MAGENTA 0xF81F
#define YELLOW 0xFFE0
#define WHITE 0xFFFF
#define AMBER 0xFDC0

// i2c GPIO-Expander Addresses - ATTENTION: Addresses only valid for PCF8574! PCF8574A requires different addresses!
#define ATTEN_ADR 0x20  // i2c-Address of GPIO-Expander that controls the Attenuator - 
#define KEYBD_ADR 0x21  // i2c Address of GPIO-Expander connected to the Keyboard - ROW_0 | ROW_1 | ROW_2 | ROW_3 | COLUMN_0 | COLUMN_1 | COLUMN_2 | FREE
#define INTGPIO_ADR 0x22  // i2c Address of the Interrupt-Receiver/GPIO-Expander - INT/GPIO-BITMAP: INT0 | INT_KEYBD | FREE | ADF4351_LOCK | AD9910_LOCK | ADF4351_CS | AD9910_CS | AD9910_DU
#define ANALOG_ADR 0x24   // i2c-Address for the Analog Board

// Other i2c-Device Addresses
#define PWR_METER_ADR 0xFF // Address of the ADC in the Power-Meter - 0xFF = Placeholder-Value

const int INT = 16;  // Interrupt-Pin: GPIO16

// System State codes - Valid System states for both the entire System, as well as the SignalSource-Modules:
const byte STATE_NORM_OP = 0x00; // Default-Screen, also called "Mode_Select_1"-Screen.
const byte STATE_INT_WAITING = 0x01;
const byte STATE_MODE_SELECT_2 = 0x02; // Second Mode-Select Screen
const byte STATE_ADV_FUNCTIONS = 0x03;
const byte STATE_ADV_MODULATION = 0x04;
const byte STATE_ARBITRARY_WAVE = 0x05;
const byte STATE_SIGPATH_CTRL = 0x06;// SignalPath-Control

const byte STATE_POWERDOWN = 0x10;
const byte STATE_COLDSTART = 0x11;
const byte STATE_STANDBY = 0x20;
const byte STATE_VCO_CHECK = 0x30;
const byte STATE_TEST = 0x40;
// Calibration and setup-Modes
const byte STATE_CALIBRATION = 0xC0;
const byte STATE_SIGSOURCE_CAL = 0xC1;
const byte STATE_SIGPATH_CAL = 0xC2;
const byte STATE_SIGPATH_COMP = 0xC3;
const byte STATE_TOUCH_CAL = 0xC4;
const byte STATE_CONFIG_1 = 0xCA; // Basic System Settings
const byte STATE_CONFIG_2 = 0xCB; 
const byte STATE_CONFIG_3 = 0xCC;
const byte STATE_CONFIG_4 = 0xCD;
const byte STATE_CONFIG_5 = 0xCE; // NV-RAM Control & Calibration
const byte STATE_CONFIG_6 = 0xCF; 
const byte STATE_DEBUG = 0xD0;
const byte STATE_DEMO = 0xD1;
// Sys-States in which the system waits for an input
const byte STATE_ENTER_FREQ = 0xE0;
const byte STATE_ENTER_AMPLITUDE = 0xE1;
const byte STATE_ENTER_MODULATION_TYPE = 0xE2;
const byte STATE_ENTER_AM_PARAMETERS = 0xE3;
const byte STATE_ENTER_FM_PARAMETERS = 0xE4;
const byte STATE_ENTER_PM_PARAMETERS = 0xE5;
const byte STATE_ENTER_RAMP = 0xEA;
const byte STATE_ENTER_BURST = 0xEB;
const byte STATE_ENTER_DIG_PATTERN = 0xEC;

// Sys-State when the device is started:
const byte STATE_INIT = 0xFF;

// InputMode Codes - 255 possible Input-Modes available
const byte INPUT_ENTER_FREQ = 0x00;
const byte INPUT_ENTER_AMPLITUDE = 0x01;
// Requesting the Parameters for the Modulation Edit-Mode.
const byte INPUT_ENTER_MODPARA1 = 0x10;
const byte INPUT_ENTER_MODPARA2 =  0x11;
const byte INPUT_ENTER_MODPARA3 = 0x12;
const byte INPUT_ENTER_MODPARA4  = 0x13;
// Requesting the Parameters for the Burst Edit-Mode.
const byte INPUT_ENTER_BURSTPARA1 = 0x20;
const byte INPUT_ENTER_BURSTPARA2 = 0x21;
const byte INPUT_ENTER_BURSTPARA3 = 0x22;
const byte INPUT_ENTER_BURSTPARA4 = 0x23;
// Requesting the Parameters for the Ramp Edit-Mode.
const byte INPUT_ENTER_RAMPPARA1 = 0x30;
const byte INPUT_ENTER_RAMPPARA2 = 0x31;
const byte INPUT_ENTER_RAMPPARA3 = 0x32;
const byte INPUT_ENTER_RAMPPARA4 = 0x33;
// Requesting the data for the digital pattern
const byte INPUT_DIG_PATTERN = 0xD0;
// Mode in which no input is expected
const byte INPUT_NO_INPUT = 0xFF;

// Modulation-Codes - Selects which modulation-Mode is active
const byte NO_MOD_SELECTED = 0x00;
const byte AM_SELECT = 0x01;
const byte FM_SELECT = 0x02;
const byte PM_SELECT = 0x03;

// System State variable - Initialized with the "ColdStart-Value" 0x11
byte SysStatus = 0x11;
byte InputMode = 0x00;

// *** Variables to configure frequency ***
unsigned long Freq = 400000000;     // Frequency-Variable in Hz - Default Value at Startup: 400MHz (ATTENTION: The ADF4351 methods require this value to be expressed in kHz!)
unsigned long Freq_Old = 400000000; // Previous selected frequency in Hz

// *** Constants and Variables to configure Analog Board ***
// DACs and ADC-Addresses
const byte PWR_SET_DAC = 0x00;
const byte AGC_BIAS_DAC = 0x00;
const byte SIGPATH_ADC = 0x48;
// ADC_Channels
const byte ADC_RF_OUT_SENSE = 0;
const byte ADC_RF_LEVEL_SENSE = 3;
// Filter-Codes
const byte BYPASS = 0x00;
const byte LPF_750 = 0x10;
const byte LPF_1500 = 0x20;
// Variables
int Atten = 95;             // Attenuation-Variable - Default: 95dB Attenuation
int OUTPUT_dBm = 0;         // Output-Power - 0 = Minimal output-amplitude. Can be adjusted in steps of 0.1 dBm. 
int OUTPUT_DBM_OFFSET = 100; // Offset-Value for the Output-Amplitude in dBm by +100, because the minimal possible output is -100dBm <-- ATTENTION! SUBJECT TO CHANGE, PENDING TESTING OF ANALOG SIGNAL CHAIN!
bool AmplitudeModeDB = true; // Show the Amplitude in dBm or mV
bool SigPathAuto = true;
bool ExtSource = false;
bool ExtAM = false;
byte LPF_Select = 0x10;

// *** Variables to configure Modulation-Parameters ***
unsigned long ModPara1 = 0; // Modulation-Strength (Amplitude, Frequency-Width, Phase-Angle
unsigned long ModPara2 = 0; // Modulation-Frequency (khz) - does not apply if digital Pattern is active
unsigned long ModPara3 = 0; // Additional Modulation-Parameter for Advanced Modulation-Mode
unsigned long ModPara4 = 0; // Additional Modulation-Parameter for Advanced Modulation-Mode
byte DigPatternData[128]; // Initialize the array to store the data for the digital pattern.
byte ModulationType = 0x00;
bool ModulationEnable = false;

// *** Variables for various tasks ***
byte interruptSource = 0; // Stores the Interrupt-Source
byte intgpio = 0xFF; // Stores the current state of the IO-Pins of the INTGPIO-Expander - Initialized with 0xFF
boolean UpdateFreqArea = true; // Shows that the Frequency-Area needs to be Updated
boolean UpdateAmpArea = true; // Shows that the Amplitude-Area needs to be Updated
boolean UpdateModArea = true; // Shows that the Modulation-Area needs to be Updated
unsigned long keyboardInput; // Stores the input made via the Keypad

void setup() {
  SysStatus = STATE_INIT;
  Serial.begin(9600);
  Serial.println(F("*** SERIAL INTERFACE - RF SIGNAL-GENERATOR ***"));

  Serial.println(F("Defining Pins"));
  pinMode(TFT_DC, OUTPUT);
  pinMode(TFT_CS, OUTPUT);

  Serial.println(F("SPI Init"));
  SPI.begin();

  Serial.println(F("i2c Init"));
  Wire.begin();

  Serial.println(F("DISPLAY-Setup"));
  tft.begin(HX8357D);
  tft.fillScreen(HX8357_BLUE);
  tft.setRotation(1);
  delay(50);
  Serial.println(F("DISPLAY-TEST"));
  TestDisplay();
  delay(0); // Delay added to yield Processor-Core to the WiFi-Code
  Serial.println(F("Show Splashscreen"));
  Splash();
  delay(2000); // Delay added to yield Processor-Core to the WiFi-Code
  tft.fillScreen(BLACK);
  tft.setTextColor(GREEN);
  tft.setCursor(10, 25);
  tft.setTextSize(2);
  tft.println(F("DISPLAY-TEST COMPLETE"));
  
  tft.println(F("Touch Init"));
  Serial.println(F("Touch Init"));
  if (! touch.begin()) {
    Serial.println(F("STMPE not found!"));
    tft.println(F("STMPE not found!"));
    ErrorMode(0x10, true);
  }

  delay(0); // Delay added to yield Processor-Core to the WiFi-Code

  Serial.println(F("Reading CAL-Data from the NV-RAM"));
  tft.println(F("Reading CAL-Data from the NV-RAM"));
  if (ReadNVData(0x5B0) == 0xFFFF) {
    Serial.println(F("SYSTEM UNCAL"));
    tft.println(F("SYSTEM UNCAL"));
    Message("UNCAL!", RED);
  }
  if (ReadNVData(0x5B0 == 0x00)) {
    Serial.println(F("No CAL-RAM installed"));
    tft.println(F("No CAL-RAM installed"));
    Message("NO CAL-MEMORY!", RED);
  }
  else {
    ReadCalData();
  }

  yield();

  Serial.println(F("Setting up analog board with default value"));
  tft.println(F("Setting up analog board with default value"));
  Wire.beginTransmission(ANALOG_ADR);
  Wire.write(B00100101); // Set SignalPathControl to Default-Value
  Wire.endTransmission();
  SetAttenuator(0x00);
  SetAmplitude(OUTPUT_dBm);
  Serial.println(F("ADS1015 INIT"));
  tft.println(F("ADS1015 INIT"));
  ads.begin();
  ads.setGain(GAIN_ONE); // Set the ADC_Gain to ONE, which means the resolution is 2mV / LSbit

  Serial.println(F("AD9910 INIT"));
  tft.println(F("AD9910 INIT"));
  SetupAD9910(STATE_INIT);
  delay(0); // Delay added to yield Processor-Core to the WiFi-Code

  Serial.println(F("ADF4351 INIT"));
  tft.println(F("ADF4351 INIT"));
  SetupADF4351(STATE_INIT);
  delay(0);

  Serial.println(F("SIGNAL PATH-TEST"));
  tft.println(F("SIGNAL PATH-TEST"));
  SignalPathTest();
  delay(1000);

  //  byte keypressed_Setup = ReadKeyboard();
  //  if (keypressed_Setup == 0xCF) {
  //    SysStatus = STATE_TOUCH_CAL;
  //    Serial.println(F("#-Key pressed - Starting Touch-Calibration"));
  //    TouchCalibration();
  //  }
  //
  //  if (keypressed_Setup == 0xFF) {
  //    SysStatus = STATE_DEBUG;
  //    Serial.println(F("*-Key pressed - Starting Touch-Calibration"));
  //  }
  //  else {
  SysStatus = STATE_NORM_OP;
  //  }
  tft.setTextSize(5);
  tft.println("READY!");
  delay(1000);

  tft.fillScreen(BLACK);
  Freq = 287234912; // Set frequency to 287.234,912 kHz to check if Calculation in Display-Function is correct
  SetFreq(Freq);
  UpdateDisplay();
  Serial.print(F("Frequency = "));
  Serial.print(Freq, DEC);
  Serial.println(F(" Hz"));
  Serial.println(F("Attaching Interrupt"));
  attachInterrupt(digitalPinToInterrupt(INT), ISR, FALLING);
}

void loop() {
  //  CheckSerial();
  switch (SysStatus) {
    case STATE_DEBUG:
      detachInterrupt(digitalPinToInterrupt(INT));
      SetFreq(60000000);
      delay(5000);
      SetFreq(50000000);
      delay(5000);
      // Put some code here to escape Debug-Mode
      attachInterrupt(digitalPinToInterrupt(INT), ISR, FALLING);
      break;
    case STATE_NORM_OP:
      detachInterrupt(digitalPinToInterrupt(INT));
      UpdateDisplay();
      attachInterrupt(digitalPinToInterrupt(INT), ISR, FALLING);
      break;
    case STATE_INT_WAITING:
      ReactInt();
      UpdateDisplay();
      attachInterrupt(digitalPinToInterrupt(INT), ISR, FALLING);
      break;
  }
  switch (ModulationEnable){
    case true:
    break;
    case false:
    break;
  }
  yield();
}

/*
   Checks if the Frequency is below or above 400MHz and then calls the Subroutines that program the signal-sources accordingly
*/
void SetFreq(unsigned long FreqSetFreq) {
  Serial.print("FRQ: ");
  Serial.print(FreqSetFreq);
  Serial.print(" Hz");
  if (keyboardInput >= 4400000000) {
    Message("ERROR: Frequency outside limits", RED);
    Freq = Freq_Old; // Set Frequency-Variable to previous value
    return; // Cancel updating Frequency
  }
  if (keyboardInput >= 3000000000) {
    Message("WARNING: Frequency outside CAL-Area", AMBER);
  }
  Freq = FreqSetFreq; // Update "Freq" in case the calling function didn't already do so - Kept in here for compatibility reasons due to some older code.
  if (FreqSetFreq >= 400000000UL) {
    Serial.println(" - Freq >= 400MHz - ");
    if (Freq_Old <= 400000000UL) {
      Serial.print(" - Freq_Old <= 400MHZ - ");
      SetupADF4351(STATE_NORM_OP); // Enable the ADF4351-Output
      Serial.print(" - ENABLE ADF4351 - ");
      SetupAD9910(STATE_STANDBY); // Disable the AD9910-Output
      Serial.print(" - AD9910 DISABLED - ");
    }
    else {

    }
    Serial.println(F(" Using ADF4351"));
    SetFreqADF4351(FreqSetFreq);
  }
  else {
    Serial.println(" - Freq <= 400MHz - ");
    if (Freq_Old >= 400000000UL) {
      Serial.print("FREQ_OLD >= 400MHZ");
      SetupAD9910(STATE_NORM_OP); // Enable the ADF4351-Output
      Serial.print(" - ENABLE AD9910 - ");
      SetupADF4351(STATE_STANDBY); // Disable the AD9910-Output
      Serial.print(" - DISABLE ADF4351 - ");
    }
    else {

    }
    Serial.println(F(" Using AD9910"));
    SetFreqAD9910(FreqSetFreq);
  }
  Freq_Old = FreqSetFreq;
  Serial.print("Freq_Old = ");
  Serial.println(Freq_Old, DEC);
  UpdateFreqArea = true;
  // Check if SigPath-Settings are set to AUTO and adjust the Filtering depending on the frequency
  switch (SigPathAuto){
    case true:
    if(Freq <= 800000000){
      SetLPF(LPF_750);
    }
    if(Freq >= 800000000 && Freq <= 1500000000){
      SetLPF(LPF_1500);
    }
    else{
      SetLPF(BYPASS);
    }
    break;
    case false:
    break;
  }
}


